/*****************************************************************************
    netio.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/param.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <limits.h>

#include <netinet/ip.h>
#include <netinet/udp.h>
#ifdef HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#include <arpa/inet.h>

#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
#include <string.h>

#include "params.h"
#include "globals.h"
#include "logfun.h"
#include "fileio.h"
#include "access.h"
#include "readconfig.h"
#include "netio.h"
#include "process.h"
#include "util.h"
#include "neterror.h"
#include "cache.h"
#include "states.h"

#ifdef MAX
#undef MAX
#endif
#define MAX(A, B)	(((A) >= (B)) ? (A) : (B))

#define RQP_NONAME      1       /* no file name */
#define RQP_DYNMEM      2       /* no dynamic memory */
#define RQP_NOMODE      3       /* no transfer mode */
#define RQP_ILLMODE     4       /* illegal/unsupported transfer mode */
#define RQP_NOOPTVAL    5       /* no option value */
#define RQP_ILLOPTVAL   6       /* illegal option value */
#define RQP_ILLFMT      7       /* illegal request format */
#define RQP_UNKOPT	8	/* unknown option */

#define IDLE_CYCLE	10	/* idle timeout cycle	*/

static int recvFile2 (tftpR_t *request, tftpD_t *recvBuffer);
static int sendFile2 (tftpR_t *request, tftpD_t *sendBuffer);

static char* tftpErrors[] =  {	NULL,
				"File not found",
				"Access violation",
				"Disk full or allocation exceeded",
				"Illegal TFTP operation",
				"Unknown transfer ID",
				"File already exists",
				"No such user"		};

static struct opl_s { int  code;
                      char *name; } 
       opcodelist[] = { {RT_RRQ,  "RRQ"},   {RT_WRQ, "WRQ"},
                        {RT_DATA, "DATA"},  {RT_ACK, "ACK"},
                        {RT_ERROR,"ERROR"}, {RT_OACK,"OACK"} };

static struct otl_s { int  otype;
                      char *oname; } 
       optionlist[] = { {ROT_BSZ, RO_BSZ}, {ROT_TMO,  RO_TMO},
                        {ROT_TSZ, RO_TSZ}, {ROT_BSZ2, RO_BSZ2} };

#define optionitems (sizeof(optionlist)/sizeof(optionlist[0]))

#ifdef  FILE_CACHE
/**/static struct timeval getSum;
#endif /* FILE_CACHE */

/*
	get rid of dynamic request structure
*/

void flushRequest(tftpR_t **request_p)
{
    tftpR_t *request = *request_p;
    if (request != NULL) {
        if (request->reqFile != NULL)
            Free(request->reqFile);
        if (request->realFile != NULL)
            Free(request->realFile);
        if (request->filter != NULL)
            Free(request->filter);
        if (request->bList != NULL)
            Free(request->bList);
        Free(request);
        *request_p = NULL;
    }
}


char *opCode2txt (int opcode)
{
    int i;
    static char buf[64];

    for (i = 0; i < (int)(sizeof(opcodelist)/sizeof(opcodelist[0])); i++)
        if (opcode == opcodelist[i].code)
            return opcodelist[i].name;

    snprintf (buf, sizeof(buf), "%d", opcode);
    return buf;
}

uint setNetBuffer (int socket, uint minLen)
{
    socklen_t vl = sizeof(uint);
    uint sbs, rbs, rawLen = minLen + sizeof(tftpH_t);

    if (getsockopt(socket, SOL_SOCKET, SO_SNDBUF, &sbs, &vl) ||
        getsockopt(socket, SOL_SOCKET, SO_RCVBUF, &rbs, &vl)) {
        LOG_MSG2 (LOG_WARNING, _("Cannot get network buffer size: %m"));
        return 0;
    }

    LOG_MSG5 (LOG_DEBUG, "Buffers send:%u recv:%u needed:%u", sbs, rbs, rawLen);

    if (sbs < rawLen) {
        if (setsockopt(socket, SOL_SOCKET, SO_SNDBUF, &rawLen, vl))
            LOG_MSG2 (LOG_WARNING, 
                      _("Cannot set send buffer size to %u: %m"), rawLen);
        getsockopt(socket, SOL_SOCKET, SO_SNDBUF, &sbs, &vl);
        LOG_MSG4 (LOG_INFO, _("Send buffer size now %u"), sbs);
    }
    if (rbs < rawLen) {
        if (setsockopt(socket, SOL_SOCKET, SO_RCVBUF, &rawLen, vl))
            LOG_MSG2 (LOG_WARNING, 
                      _("Cannot set recv buffer size to %u: %m"), rawLen);
        getsockopt(socket, SOL_SOCKET, SO_RCVBUF, &rbs, &vl);
        LOG_MSG4 (LOG_INFO, _("Recv buffer size now %u"), rbs);
    }
    return (MIN(rbs, sbs));
}


int setBits(in_addr_t mask)
{
    int setbits, allbits;

    for (setbits=0, allbits=(8*sizeof(in_addr_t)); allbits; allbits--) {
        if (mask & 1)
            setbits++;
        mask >>= 1;
    }
    return setbits;
}


void presetRequest (tftpR_t *request, int role, sopair_t *sockets)
{
    memset(request, 0, sizeof(tftpR_t));

    request->reqRole           = role;
    request->option.blocksize  = SEGSIZE;
    request->times             = defTimer;
    if (sockets != NULL)
        request->reqSock       = *sockets;
}    


char *getErrText (tftpD_t *buf, int buflen)
{
    int i;

/*      we accept printable characters only     */

    for (i = 0; i < (int)(buflen-sizeof(tftpH_t)); i++)
	if (!isprint(buf->data[i]))
	    break;

/*      did we get a string terminator at last ?
	if so, return the error string from the packet
*/
    if (!i)
        return NULL;

    if (buf->data[i] == '\0')
	return buf->data;

/*      nothing valid in there
	try to deduce the text from the error code
*/
    switch ((i = (int)ntohs(buf->head.errCode))) {
	case PE_NOTFOUND:
	case PE_ACCESS:
	case PE_NOSPACE:
	case PE_BADOP:
	case PE_BADID:
	case PE_EXISTS:
	case PE_NOUSER:
	    return tftpErrors[i];
	default:
	    break;
    }

/*      nothing valid - return NULL */

    return NULL;
}


/*
	compute the elapsed in seconds time from the runtime statistics
*/

double elapTime (tftpR_t *request)
{
    return (tv2double(&request->times.stop)-tv2double(&request->times.start));
}

/*
	compute the xfer-rate in kBytes/sec from the runtime statistics.
*/

double xfrRate (tftpR_t *request)
{
    return  ((double)request->iostat)/elapTime(request)/1000.0;
}

/*	do runtime status trace	*/

void statusTrace(tftpR_t *rq)
{
    char *pfx1, *pfx2;
    char* fmt1 = "Local I/O (Disk) %ld Bytes %s";
    char* fmt2 = "Network I/O %ld Bytes %s in %gs (%g kBytes/s)";

    rq->times.stop = uftime(NULL);

    if (rq->reqType == RT_WRQ) {
        pfx1 = "written";
        pfx2 = "received";
    }
    else {
        pfx1 = "read";
        pfx2 = "transmitted";
    }

#ifdef  FILE_CACHE
    if (rq->fcache == NULL)
#endif /* FILE_CACHE */
        LOG_MSG4 (LOG_INFO, fmt1, rq->iostat, pfx1);

    LOG_MSG3 (LOG_NOTICE, fmt2, rq->xfrstat, pfx2, elapTime(rq), xfrRate(rq));
    return;
}

/*
	compare requested options with OACKed options
*/

int checkOptions(char *data, int len, tftpR_t *req)
{
    int i, rtc=0;
    opt_t option;

    if (parseOptions(data, len, &option))
        return 1;

    for (i = 0; i < (int)(optionitems); i++) {
        if (req->option.types & optionlist[i].otype) {
            if (option.types & optionlist[i].otype) {
                switch (optionlist[i].otype) {
#ifdef BLOCKSIZE_2
                    case ROT_BSZ2:
                        LOG_MSG5 (LOG_DEBUG, "Blocksize2 Req:%d, Ack:%d",
                                  req->option.blocksize, option.blocksize);
                        if ((req->option.mandatory > 1) &&
                            (powerOf2(req->option.blocksize) != 
                             option.blocksize)) {
                            LOG_MSG1 (LOG_ERR, _("Mandatory option '%s' not acceptable OACKed"), optionlist[i].oname);
                            rtc = 1;
                        }
                        req->option.blocksize = option.blocksize;
                        break;
#endif /* BLOCKSIZE_2 */
                    case ROT_BSZ:
                        LOG_MSG5 (LOG_DEBUG, "Blocksize Req:%d, Ack:%d",
                                  req->option.blocksize, option.blocksize);
                        if ((req->option.mandatory > 1) &&
                            (req->option.blocksize != option.blocksize)) {
                            LOG_MSG1 (LOG_ERR, _("Mandatory option '%s' not unchaged OACKed"), optionlist[i].oname);
                            rtc = 1;
                        }
                        req->option.blocksize = option.blocksize;
                        break;
                    case ROT_TMO:
                        LOG_MSG5 (LOG_DEBUG, "Timeout Req:%d, Ack:%d",
                                  req->option.timeout.tv_sec, 
                                  option.timeout.tv_sec);
                        req->option.timeout = option.timeout;
                        break;
                    case ROT_TSZ:
                        LOG_MSG5 (LOG_DEBUG, "Totalsize Req:%u, Ack:%u",
                                  req->option.totalsize, option.totalsize);
                        req->option.totalsize = option.totalsize;
                        break;
                }
            }
            else {
                if (req->option.mandatory) {
                    LOG_MSG3 (LOG_NOTICE, _("Mandatory option '%s' not OACKed"), 
                              optionlist[i].oname);
                    rtc  = 1;
                }
                LOG_MSG3 (LOG_NOTICE, _("Option '%s' not OACKed"), 
                          optionlist[i].oname);
                switch (optionlist[i].otype) {
#ifdef BLOCKSIZE_2
                    case ROT_BSZ2:
#endif /* BLOCKSIZE_2 */
                    case ROT_BSZ:
                        req->option.blocksize = SEGSIZE;
                        break;
                    case ROT_TMO:
                        break;
                    case ROT_TSZ:
                        break;
                }
            }
        }
        else {
            if (option.types & optionlist[i].otype) {
                LOG_MSG3 (LOG_NOTICE, _("Unrequested option '%s' OACKed"), 
                          optionlist[i].oname);
                rtc = 1;
            }
        }
    }
    return rtc;
}

/*
	extract options from a text field. 
	don't rely on the format of the protocol description
*/

int parseOptions (char *reqtext, int len, opt_t *op)
{
    char *start, *option = "";
    int i, n, state, otype = 0, rtc = 0;

    op->types = 0;

    for (i = 0, state = 1, start = reqtext; i < len; i++, reqtext++) {
	if (*reqtext == 0) {
	    switch (state = state ? 0 : 1) {
                case 0:		/* options */
		    if (reqtext == start)
                        break;
                    option = start;
                    otype  = 0;
                    for (n=0; n<(int)(optionitems); n++) {
                        if (!strcasecmp(start, optionlist[n].oname)) {
                            option = optionlist[n].oname;
                            otype  = optionlist[n].otype;
                            break;
                        }
                    }
                    if (!otype)
                        LOG_MSG2 (LOG_WARNING, _("Unsupported option '%s'"),
                                  start);
                    start  = reqtext + 1;
                    break;
                case 1:		/* option value */
		    if (reqtext == start) {
                        LOG_MSG1 (LOG_ERR, _("No value for option '%s'"), 
                                  option);
			rtc = RQP_NOOPTVAL;
                    }
                    switch (otype) {
#ifdef BLOCKSIZE_2
                        case ROT_BSZ2:
                            if (sscanf(start,"%d", &op->blocksize)==1) {
                                int p2size;
                                p2size = powerOf2(op->blocksize);
                                if ((p2size >= RO_BSZ2MIN)  &&
                                    (p2size <= RO_BSZ2MAX)) {
                                    op->types |= ROT_BSZ2;
                                    LOG_MSG6 (LOG_DEBUG, 
                                              "%s: 'blksize2 %d' -> %d",
                                              __FUNCTION__, 
                                              op->blocksize, p2size);
                                    op->blocksize = p2size;
                                    break;
                                }
                            }
                            LOG_MSG1 (LOG_ERR,
                                      _("Illegal option value 'blksize2 %s'"),
                                      start);
                            rtc = RQP_ILLOPTVAL;
                            break;
#endif /* BLOCKSIZE_2 */
                        case ROT_BSZ:
                            if (sscanf(start,"%d", &op->blocksize)==1) {
                                if ((op->blocksize >= RO_BSZMIN)  &&
                                    (op->blocksize <= RO_BSZMAX)) {
                                    op->types |= ROT_BSZ;
                                    LOG_MSG6 (LOG_DEBUG, "%s: 'blksize %d'",
                                              __FUNCTION__, op->blocksize);
                                    break;
                                }
                            }
                            LOG_MSG1 (LOG_ERR,
                                      _("Illegal option value 'blksize %s'"),
                                      start);
                            rtc = RQP_ILLOPTVAL;
                            break;

                        case ROT_TMO:
                            if (sscanf(start,"%ld", &op->timeout.tv_sec) == 1) {
                                if ((op->timeout.tv_sec >= RO_TMOMIN)  &&
                                    (op->timeout.tv_sec <= RO_TMOMAX)) {
                                    op->timeout.tv_usec = 0;
                                    op->types |= ROT_TMO;
                                    LOG_MSG6 (LOG_DEBUG, "%s: 'timeout %ds'",
                                              __FUNCTION__,op->timeout.tv_sec);
                                    break;
                                }
                            }
                            LOG_MSG1 (LOG_ERR,
                                      _("Illegal option value 'timeout %s'"),
                                      start);
                            rtc = RQP_ILLOPTVAL;
                            break;

                        case ROT_TSZ:
                            if (sscanf(start,"%u", &op->totalsize) == 1) {
                                if (op->totalsize <= RO_TSZMAX) {
                                    op->types |= ROT_TSZ;
                                    LOG_MSG6 (LOG_DEBUG, "%s: 'tsize %u'",
                                              __FUNCTION__, op->totalsize);
                                    break;
                                }
                            }
                            LOG_MSG1 (LOG_ERR,
                                      _("Illegal option value 'tsize %s'"),
                                      start);
                            rtc = RQP_ILLOPTVAL;
                            break;

                        default:
                            LOG_MSG3 (LOG_NOTICE, 
                                      _("Unsupported Option '%s %s'"),
                                      option, start);
                            rtc = RQP_UNKOPT;
                    }
                    start  = reqtext + 1;
                    state = 0;
                    break;
	    }
	}
    }
    if (!op->types) {
        LOG_MSG6 (LOG_DEBUG, "%s: no options", __FUNCTION__);
    }
    return rtc;
}

/*
	extract file name, transfer mode strings and options from the
	text field of the initial request packet.
	don't rely on the format of the protocol description
*/

static int parseRequest (char *reqtext, int len, tftpR_t *req)
{
    char *start;
    int i, state, rtc=0;

    LOG_MSG6 (LOG_DEBUG, "%s: %d bytes", __FUNCTION__, len);

    for (i = 0, state = 0, start = reqtext; i < len; i++, reqtext++) {
	if (*reqtext == 0) {
	    switch (++state) {
		case 1:		/* file name */
		    if (reqtext == start) {
                        LOG_MSG1 (LOG_ERR, _("No file name in request"));
			return RQP_NONAME;
                    }
		    if ((req->reqFile = Strdup(start)) == NULL) {
                        errNoDynmem;
                        return RQP_DYNMEM;
                    }
                    LOG_MSG6 (LOG_DEBUG, "%s: filename '%s'", __FUNCTION__,
                              req->reqFile);
		    start = reqtext+1;
		    break;
		case 2:		/* transfer mode */
		    if (reqtext == start) {
                        LOG_MSG1 (LOG_ERR, _("No transfer mode in request"));
			return RQP_NOMODE;
                    }
                    if (!strcasecmp(RMS_OCTET, start))
                        req->reqMode = RM_OCTET;
                    else if (!strcmp(RMS_NETASCII, start))
                        req->reqMode = RM_NETASCII;
                    else {
                        LOG_MSG1 (LOG_ERR, 
                                  _("Unsupported/unknown transfer mode '%s'"),
                                  start);
                        return RQP_ILLMODE;
                    }
                    LOG_MSG6 (LOG_DEBUG, "%s: transfer mode '%d'", 
                              __FUNCTION__, req->reqMode);
                    
                    rtc = parseOptions (reqtext+1, len-i-1, &req->option);
                    if (!xferOptions && req->option.types) {
                        LOG_MSG4 (LOG_INFO, _("Requested options ignored"));
                        req->option.types = 0;
                    }
                    return rtc;
	    }
	}
    }
    return (state >= 2) ? 0 : RQP_ILLFMT;
}

/*
	append an option list to a buffer
*/

char *appendOptions (char *buffer, tftpR_t *req, int *len)
{
    static char valbuf[64];
    int i, dyn=0, olen=0, bufsiz=128, nlen, vlen;
    
    if (buffer == NULL) {
        if ((buffer = Malloc(bufsiz)) == NULL) {
            errNoDynmem;
            *len = -1;
            return NULL;
        }
        dyn=1;
    }
    else {
        bufsiz = *len;
    }

    *len = -1;

    for (i = 0; i < (int)(optionitems); i++) {
        if (req->option.types & optionlist[i].otype) {
            *valbuf = '\0';
            switch (optionlist[i].otype) {
#ifdef BLOCKSIZE_2
                case ROT_BSZ2:
                    snprintf(valbuf, sizeof(valbuf), "%d", 
                             req->option.blocksize);
                    break;
#endif /* BLOCKSIZE_2 */
                case ROT_BSZ:
                    snprintf(valbuf, sizeof(valbuf), "%d", 
                             req->option.blocksize);
                    break;

                case ROT_TMO:
                    snprintf(valbuf, sizeof(valbuf), "%ld", 
                             req->option.timeout.tv_sec);
                    break;

                case ROT_TSZ:
                    snprintf(valbuf, sizeof(valbuf), "%u", 
                             req->option.totalsize);
                    break;
            }
            if (*valbuf == '\0') {
                LOG_MSG1 (LOG_ERR, _("Error converting option '%s' value"),
                          optionlist[i].oname);
                if (dyn)
                    Free(buffer);
                return NULL;
            }
            nlen = strlen(optionlist[i].oname) + 1;
            vlen = strlen(valbuf) + 1;
            if ((olen+nlen+vlen) > bufsiz) {
                LOG_MSG1 (LOG_ERR, _("Options too long (%d)"), olen+nlen+vlen);
                if (dyn)
                    Free(buffer);
                return NULL;
            }
            strcpy(&buffer[olen], optionlist[i].oname);
            olen += nlen;
            strcpy(&buffer[olen], valbuf);
            olen += vlen;
            LOG_MSG6 (LOG_DEBUG, "Option appended '%s %s'", 
                      optionlist[i].oname, valbuf);
        }
    }
    *len = olen;
    return buffer;        
}

/*
	format current options into atext string
*/

char *textOptions (tftpR_t *req)
{
    static char optbuf[100];
    int i, len=0;

    optbuf[0] = '\0';

    for (i = 0; i < (int)(optionitems); i++) {
        if (req->option.types & optionlist[i].otype) {
            if (len) {
                strcpy(&optbuf[len], ", ");
                len += 2;
            }
            switch (optionlist[i].otype) {
#ifdef BLOCKSIZE_2
                case ROT_BSZ2:
#endif /* BLOCKSIZE_2 */
                case ROT_BSZ:
                    snprintf(&optbuf[len], sizeof(optbuf) - len, "%s=%d", 
                            optionlist[i].oname, req->option.blocksize);
                    break;
                case ROT_TMO:
                    snprintf(&optbuf[len], sizeof(optbuf) - len, "%s=%ld", 
                            optionlist[i].oname, req->option.timeout.tv_sec);
                    break;
                case ROT_TSZ:
                    snprintf(&optbuf[len], sizeof(optbuf) - len, "%s=%u", 
                            optionlist[i].oname, req->option.totalsize);
                    break;
            }
            len = strlen(optbuf);
        }
    }
    return optbuf;
}

/*
	convert a dns/host name or dottet decimal ip address
	into binary representation.
	returns 0 on success;
*/

int name2sai (char *host, char *cport, sai_t *saip)
{
    in_addr_t addr;
    int port;

    addr =  ((in_addr_t)inet_addr(host));
 
    if ((addr == (in_addr_t)INADDR_NONE)) {
	struct hostent *hp;
	hp = gethostbyname(host);
	if (hp == NULL)  {
	    LOG_MSG1 (LOG_ERR, _("Illegal host name '%s'"), host);
	    return 1;
	}
	addr = *((in_addr_t *)hp->h_addr);
    }

    if ((port = setPort(cport)) <= 0) {
	LOG_MSG1 (LOG_ERR, _("Illegal port '%s'"), cport);
	return 1;
    }

    saip->sin_family      = AF_INET;
    saip->sin_addr.s_addr = addr;
    saip->sin_port        = htons((short)port);

    return 0;
}


in_addr_t xinet_addr (char *host, int resolv)
{
    in_addr_t addr;

    if (host == NULL)
        return INADDR_ANY;

    addr =  ((in_addr_t)inet_addr(host));

    if ((addr == (in_addr_t)INADDR_NONE) && resolv) {
        struct hostent *hp;
        hp = gethostbyname(host);
        if (hp != NULL)
            addr = *((in_addr_t *)hp->h_addr);
    }

    return addr;
}

/*
	basically same as the library function inet_ntoa
	but tries to resolve the host address into a host name
	if not prohibited by 'resolv'
*/

char *xinet_ntoa (in_addr_t ia, int resolv)
{
    struct hostent *he;
    struct in_addr sia;

    if (!ia)
        return "0.0.0.0";

    if (resolv) {
	he = gethostbyaddr ((char *) &ia, sizeof (in_addr_t), AF_INET);
	if (he != NULL)
	    return he->h_name;
    }

    sia.s_addr = ia;
    return inet_ntoa(sia);
}

/*
	basically same as the function xinet_ntoa used for converting net masks
	does no resolving, but tries to convert the netmask to CIDR notation
	if possible
*/

char *xmask_ntoa (in_addr_t ia)
{
    unsigned long reg;
    int i;
    char *result;
    static char buf[16];

    for (reg = (ulong)htonl(ia), i = 0; reg; reg >>= 1) {
        if (i) {
            if (!(reg & 1)) {
                i = -1;
                break;
            }
            i++;
        }
        else if ((reg & 1))
            i = 1;
    }

    if ((i >= 0) && (i <= 128)) {
        snprintf(buf, sizeof(buf), "/ %d", i);
        result = buf;
    }
    else
        result = xinet_ntoa(ia, 0);

    return result;
}


int setPort (char *cPort)
{
    struct servent *serv;
    int port = -1;

    if (cPort != NULL) {

        sscanf(cPort, "%d", &port);

        if (port >= 0)
            return port;

        if ((serv = getservbyname (cPort, "udp")) != NULL)
            return (int)ntohs(serv->s_port);
    }

    return -1;
}		   


/*
	check if this file descriptor refers to a socket
	used to determine if we're forked by inetd or started standalone
*/

int isaSocketFd (int fd)
{
    int dummy;
    socklen_t length = sizeof(int);

    return (!getsockopt(fd, SOL_SOCKET, SO_TYPE, &dummy, &length));
}

/*
	check if this file pointer refers to a socket
	used to determine if we're forked by inetd or started standalone
*/

int isaSocket (FILE *fp)
{
    return (isaSocketFd(fileno(fp)));
}

/*
	send an error message
*/

int sendToError (int socket, sai_t *peer, int err, char *errText, tout_t *times)
{
    tftpD_t errBuffer;
    struct timeval timeout;
    fd_set fdSel;
    int ns, ios, try;

    if (times == NULL)
        times = &defTimer;

    errBuffer.head.opCode  = htons(RT_ERROR);
    errBuffer.head.errCode = htons(((int)err));
    errBuffer.data[0]      = 0;

    switch (err) {
	case PE_INTERNAL:
	    err = PE_UNDEF;
	    if (errText == NULL)
		errText = "Internal server error";
	    strncpy(errBuffer.data, errText, SEGSIZE);
	    break;
	case PE_SYSTEM:
	    err = PE_UNDEF;
	    if (errText == NULL)
		errText = strerror(errno);
	    strncpy(errBuffer.data, errText, SEGSIZE);
	    break;
	case PE_UNDEF:
	    if  (errText != NULL) 
		strncpy(errBuffer.data, errText, SEGSIZE);
	    break;
	case PE_NOTFOUND:
	case PE_ACCESS:
	case PE_NOSPACE:
	case PE_BADOP:
	case PE_BADID:
	case PE_EXISTS:
	case PE_NOUSER:
	    strncpy(errBuffer.data, tftpErrors[err], SEGSIZE);
	    break;
	default:
	    return -1;
    }

    LOG_MSG4 (LOG_INFO, _("Sending error %d to %s: %s"), 
              err, xinet_ntoa(peer->sin_addr.s_addr, doResolv), errBuffer.data);

    ns = 1 + socket;

    errBuffer.head.errCode = htons(err);
    errBuffer.data[SEGSIZE-1] = 0;

    for (try = 0; try < 3; try++) {

        FD_ZERO(&fdSel);
        FD_SET(socket, &fdSel);

        timeout = proBackoff(times, EBS_SEND, 1);

        ios = select (ns, NULL, &fdSel, NULL, &timeout);

        if ((ios >= 0) || (errno != EINTR))
            break;
    }

    if (ios == 0) {
        logSelTimeout("ERR/MSG", 1, &times->send);
	return -1;
    }

    if (ios <= 0) {
        logSelError("ERR/MSG");
	return -1;
    }

    if (sendto(socket, &errBuffer, sizeof(tftpH_t)+strlen(errBuffer.data)+1, 
               0, (sa_t *)peer, sizeof(sai_t))) {
	LOG_MSG1 (LOG_ERR, _("Cannot send error message: %m"));
	return -1;
    }
    return 0;
}

/*
	send an error message to a peer
*/

int sendPeerError (tftpR_t *request, sai_t *peer, int err, char *errText)
{
    return (sendToError(request->Ssock, peer, err, errText, &request->times));
}
/*
	send an error message to the current peer
*/

int sendError (tftpR_t *request, int err, char *errText)
{
    return (sendToError(request->Ssock, &request->reqPeer, err, errText, 
                        &request->times));
}

/*
	perform the initial network startup
	this function is generally called only once at main stream startup
*/

sopair_t *netStart (in_addr_t addr, int port)
{
    int psoc;
    socklen_t lsalen;
    sai_t lsa;
    static sopair_t sockets;

    if (port < 0) {
	LOG_MSG1 (LOG_ERR, _("Illegal server port %d"), port);
        return NULL;
    }

    if ((psoc = socket (PF_INET, SOCK_DGRAM, 0)) < 0) {
	LOG_MSG1 (LOG_ERR, _("Cannot obtain primary socket: %m"));
	return NULL;
    }

    if (broadCast) {
	int on = 1;
	if (setsockopt(psoc, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on))) {
	    LOG_MSG1 (LOG_ERR, _("Cannot enable broadcast on primary socket: %m"));
	    close (psoc);
	    return NULL;
	}
	LOG_MSG4 (LOG_INFO, _("Broadcast receiver enabled"));
    }

    lsa.sin_family      = AF_INET;
    lsa.sin_addr.s_addr = addr;
    lsa.sin_port        = htons((short)port);

    if (bind (psoc, (sa_t *) &lsa, sizeof(lsa)) < 0) {
	LOG_MSG1 (LOG_ERR, _("Cannot bind to primary socket: %m"));
	close (psoc);
	return NULL;
    }

    lsalen = sizeof(sai_t);
    if (getsockname(psoc, (sa_t *) &lsa, &lsalen)) {
        LOG_MSG1 (LOG_ERR, _("Cannot get name of primary socket"));
        close (psoc);
        return NULL;
    }
    else {
        LOG_MSG5 (LOG_DEBUG, "Bound to %s:%hd",
                  xinet_ntoa(lsa.sin_addr.s_addr, doResolv), 
                  ntohs(lsa.sin_port));
    }

    sockets.send = sockets.recv = psoc;

    return &sockets;
}

int getRequest (sopair_t *sockets, tftpR_t **request_p)
{
    int recvLength;			/* received data in bytes	*/
    int rtc;				/* general return code		*/ 
    fd_set fdSel;			/* select bitmask		*/
    struct timeval timeout;		/* needed for non-timed select	*/
    struct timeval timer;		/* guess what !			*/
    static tftpD_t recvBuffer;		/*	receive Buffer		*/
    static tftpR_t requestTemplate;	/* request structure template	*/
    static tftpR_t *request;		/* ptr to requestTemplate	*/
    socklen_t      peerlen;		/* length of socket address	*/
    int maxsock;

    *request_p = NULL;
    request = &requestTemplate;
    presetRequest(request, RR_SERVER, sockets);

    startTimer(&timer);

    for (;;) {
        timeout = request->times.conn;

        LOG_MSG6 (LOG_DEBUG, "Waiting for client request, timeout %.1fs",
                  tv2double(&timeout));

	FD_ZERO(&fdSel);
	FD_SET(sockets->recv, &fdSel);
        maxsock = sockets->recv + 1;

 	rtc = select (maxsock, &fdSel, NULL, NULL, &timeout);

        switch(checkSelError(request, rtc, BM_GRQ)) {
            case PC_ERROR:
                return -1;
            case PC_AGAIN:	/* caused by signal etc. */
                return 0;
            case PC_RETRY:	/* timeout, reconfig- or dump-request */
	        if (checkTimer(timer, request->times.conn) ||
                    forceConfig || doDump)
                    return 0;
                continue;
        }

        if (!(FD_ISSET(sockets->recv, &fdSel)))
            continue;

/*	probably an incoming request 
	we do not honor new transfer requests as long as 
	a reconfiguration or a dump is pending
*/

        if (forceConfig || doDump)
            return 0;

        request->reqPeer.sin_addr.s_addr = INADDR_ANY;
        request->reqPeer.sin_port        = 0;
        peerlen = sizeof(sai_t);

	recvLength = recvfrom(sockets->recv, &recvBuffer, sizeof(recvBuffer),
			      MSG_TRUNC, (sai_t *)&request->reqPeer, &peerlen);

/*	check for bandwith limitation
*/
        if (!checkLimits(request)) 	/* max.# threads running? */
            continue;

/*	parse current request 
*/
        switch(checkGrq(request, &recvBuffer, recvLength)){
            case PC_ERROR:
	        return -1;
            case PC_AGAIN:	/* caused by seignal etc. */
                return 0;
            case PC_RETRY:
                continue;
            case PC_UNKP:
		LOG_MSG1 (LOG_ERR, _("Illegal transfer start code %d by %s"),
			 request->reqType, 
			 xinet_ntoa((in_addr_t)request->reqPeer.sin_addr.s_addr,
			 doResolv & RES_RUNTIME));
		sendError(request, PE_BADOP, NULL);
	        continue;
        }

	request->reqType    = ntohs(recvBuffer.head.opCode);

	rtc = parseRequest (recvBuffer.head.reqText, 
			    recvLength - sizeof(recvBuffer.head.opCode),
			    request);

	if (!rtc) 	/* everything o.k., shoot up transfer	*/
            break;

        switch (rtc) {
	  case RQP_ILLMODE:
            sendError(request, PE_UNDEF, "Unsupported/unknown transfer mode");
	    break;
          case RQP_DYNMEM:
	    sendError(request, PE_NOSPACE, NULL);
            break;
          case RQP_NOOPTVAL:
          case RQP_ILLOPTVAL:
	    sendError(request, PE_ILLOPTVAL, NULL);
	    break;
          default:
	    sendError(request, PE_BADOP, NULL);
	    break;
        }
    }		/* for (;;) end here	*/


    LOG_MSG3 (LOG_NOTICE, _("%s %srequest from %s for %s"),
	      request->reqMode == RM_OCTET ? _("Binary") : _("ASCII"),
	      request->reqType == RT_RRQ ? _("Read ") : _("Write "), 
	      xinet_ntoa((in_addr_t)request->reqPeer.sin_addr.s_addr,
			 doResolv & RES_RUNTIME),
	      request->reqFile);

    if (request->option.types)
        LOG_MSG3 (LOG_NOTICE, _("Options requested: %s"),
                  textOptions(request));

    if ((rtc = validateACL(request))) {
        LOG_MSG5(LOG_DEBUG, _("Access check stage 1 (ACL) failed"));
        if (!silentDeny && !(rtc & PX_DISCARD))
            sendError(request, (rtc & PE_ERRMASK), NULL);
        return 0;
    }

    if ((*request_p = (tftpR_t *)Malloc(sizeof(tftpR_t))) == NULL) {
        errNoDynmem;
        return -1;
    }

    memcpy (*request_p, request, sizeof(tftpR_t));

    return 1;
}

/*
	set up a new connection to the current peer
	this function is called after a successful 'getRequest()',
	when a new child process has been created
*/

int connectPeer (tftpR_t *request)
{
    int newSocket;
    sai_t lsa;
    uint bstat;

    if ((newSocket = socket (PF_INET, SOCK_DGRAM, 0)) < 0) {
	LOG_MSG1 (LOG_ERR, _("Cannot obtain slave socket: %m"));
	return -1;
    }

    lsa.sin_family      = AF_INET;
    lsa.sin_addr.s_addr = INADDR_ANY;
    lsa.sin_port        = 0;

    if (bind(newSocket, (sa_t *)&lsa, sizeof(lsa)) < 0) {
	LOG_MSG1 (LOG_ERR, _("Cannot bind to slave socket: %m"));
	close (newSocket);
	return -1;
    }

    if (connect(newSocket, (sa_t *)&request->reqPeer, 
                sizeof(request->reqPeer)) < 0) {
	LOG_MSG1 (LOG_ERR, _("Cannot connect to slave socket: %m"));
	close (newSocket);
	return -1;
    } 

    request->Rsock = request->Ssock = newSocket;
    request->xfrstat = request->iostat = 0;

/*
	let's check if the current block size fits the rec/xmit buffer size
        if not, reduce the block zize apropriately
*/
    bstat = setNetBuffer(newSocket, request->option.blocksize);
    if ((bstat > 0) &&  (bstat < (uint)request->option.blocksize) &&
              (request->option.types & (ROT_BSZ | ROT_BSZ2))) {
        request->option.blocksize = bstat;
        LOG_MSG3 (LOG_NOTICE, _("Reducing requested block size to %u"), bstat);
    }

    if (msgLevel >= LVL_DEBUG) {
	sai_t localSock;
	socklen_t sockLength = sizeof(localSock);
	if (getsockname(request->Rsock, (sa_t *)&localSock, &sockLength))
	    localSock.sin_port = 0;
	LOG_MSG4 (LOG_INFO, _("Local port %hd connected to %s:%hd"),
		  ntohs(localSock.sin_port),
		  xinet_ntoa((in_addr_t)request->reqPeer.sin_addr.s_addr,
			     doResolv & RES_RUNTIME),
		  ntohs(request->reqPeer.sin_port));
    }

    return 0;
}

    
/*
	write request (WRQ): receive a file from the client
	we assume an already opened file and an active socket
	all security checks  re done and 
	options (if applicable) are processed
	mode may be one of RM_OCTET or RM_NETASCII

	jacket routine to recvFile2 (see below)
	this is done to assure an easy cleanup of dyn memory
	at function return
*/

int recvFile (tftpR_t *request)
{
    tftpD_t *recvBuffer;	/* receive buffer (one packet)	*/
    int ios, rtc, bufsize;

    if ((request->option.types & ROT_TMO)) {
        request->times.recv = request->times.sack = request->option.timeout;
        LOG_MSG4 (LOG_INFO, _("New Timing: %s"), 
                             timer2asc(&request->times, T2A_XFER|T2A_RET));
    }
    LOG_MSG5 (LOG_DEBUG, _("Progressive backoff: %s"), 
                           listBackoff(&request->times, EBS_RECV));

    request->times.start = request->times.stop = uftime(NULL);

    LOG_MSG5 (LOG_DEBUG, "Blocksize: %d %s", request->option.blocksize,
              (request->option.types & (ROT_BSZ | ROT_BSZ2)) ? 
                  "(requested by client)" : "");

    bufsize = MAX(request->option.blocksize, SEGSIZE) + sizeof(tftpH_t);

    if ((recvBuffer = (tftpD_t *)Malloc(bufsize)) == NULL) {
        errNoDynmem;
        request->times.stop = uftime(NULL);
        return -1;
    }

    ios = initXferBuffer(request->option.totalsize);

    if (ios) {
        sendError(request, PE_NOSPACE, NULL);
        return -1;
    }

    rtc = recvFile2 (request, recvBuffer);

    rmXferBuffer();

    request->times.stop = uftime(NULL);
    Free(recvBuffer);
    return rtc;
}

/*
	transfer processing according to RFC 2347:

	client                                           server
	-------------------------------------------------------
	|2|barfile|0|octet|0|blksize|0|2048|0|  -->         WRQ
	                       <--  |6|blksize|0|2048|0|   OACK
	|3|1| 2048 octets of data |  -->                   DATA
	                                       <--  |4|1|   ACK
	|3|2| 2048 octets of data |  -->                   DATA
	                                       <--  |4|2|   ACK
	|3|3|<2048 octets of data |  -->                   DATA
*/

static int recvFile2 (tftpR_t *request, tftpD_t *recvBuffer)
{
    tftpD_t ackBuffer;		/* acknowledge packet		*/
    int recvLength = 0;		/* received bytes		*/
    int hsock;			/* highest socket+1 (select)	*/
    ushort lastDat = 0;		/* latest block ACK'ed		*/
    int done = 0;		/* flag for end-of-transfer	*/
    int rtc;			/* general function return code	*/
    fd_set fdSel;		/* select bitmap		*/
    int plen;			/* ACK packet length		*/
    enum io_states state;	/* state machine status		*/
    struct timeval timeout;	/* i/o timer			*/

    hsock = MAX(request->Rsock, request->Ssock) + 1;
    request->curBlock = 1;		/* next block awaited		*/

/*
	Loop through the sate-machine that implements the receive algorithm
*/

    state = (xferOptions && request->option.types) ? ST_OACKWAIT : ST_ACKWAIT;

    for (;;) {

        if (doTrace) {
            doTrace = 0;
            statusTrace (request);
        }

	LOG_MSG6 (LOG_DEBUG, "%s: state %d: %s",
                  __FUNCTION__, state,state_names[state]);

	switch (state) {

            default:
                LOG_MSG1 (LOG_ERR, _("Internal state machine error"));
                break;
/*
	wait for data packet from peer
	this is actually a 'wait for any reply',
	because we shoud be prepared to receive an packet,
	especially error packets.

	On timeout, we check the retry-count.
	This counter will be reset on any sucessfully transferred packet.
	If retries are still allowed, we re-ack the latest packet.
*/
	    case ST_RECVWAIT:
		FD_ZERO(&fdSel);
		FD_SET(request->Rsock, &fdSel);

		timeout = proBackoff(&request->times, EBS_RECV, request->retry);

                LOG_MSG6 (LOG_DEBUG, "%s: timeout %.1fs", __FUNCTION__,
                          tv2double(&timeout));

		rtc = select (hsock, &fdSel, NULL, NULL, &timeout);

                switch(checkSelError(request, rtc, BM_DATA)) {
                    case PC_ERROR:
                        return -1;
                    case PC_AGAIN:
                        continue;
                    case PC_RETRY:
                        state = ST_ACKWAIT;
                        break;
                }
		state = ST_RECV;
		break;
/*
	The received packed should be at least of length sizeof(tftpD_t)
	(usually 4 bytes). Short packets are silently discarded:
	we proceed as with timeouts above.

	Now let's check the opcode of the received packets.
	All non-data packets are considered fatal transfer errors, which
	lead to an abort of the current transfer.
*/
	    case ST_RECV:
		recvLength = recv(request->Rsock, recvBuffer, 
				  request->option.blocksize+sizeof(tftpH_t),
                                  MSG_TRUNC);

		LOG_MSG6 (LOG_DEBUG, "%s: %d/%d bytes received",
                          __FUNCTION__, recvLength, recvLength-sizeof(tftpH_t));

                switch (checkData(request, recvBuffer, recvLength)) {
                    case PC_ERROR:
                        return -1;
                    case PC_RETRY:
                        state = ST_ACKWAIT;
                        break;
                }
/*
	We got the desired data packet, so let's check the block number.
	Only 2 cases are valid: (current-block) and (current-block -1),

	If we receive (current-block -1) we first check the retry count.
	If retries are still allowed, we simply acknowledge the packet
	and discard the data, because this is apparently a re-sent packet.
*/
                lastDat = ntohs(recvBuffer->head.blockNo);

		if (lastDat == hs(request->curBlock)) {
		    state = ST_PUTDATA;
		    break;
                }

		if (lastDat == (hs(request->curBlock-1))) {
                    logRepeatRequest("DATA", (int)lastDat);
		}
                else {
                    logSequenceError("DATA", (int)lastDat, 
                                     (int)hs(request->curBlock));
		}
                if (retryPossible(request, PE_BADOP, "DATA out of sequence")
                    == PC_ERROR)
                    return -1;
                state = ST_ACKWAIT;
		break;
/*
	Now, this is what we expect to come in next.
        Check for file size limit, then
	let's look at the length: anything  < SEGSIZE is a 'final packet'.
	in this case prepare for closing the transfer.
*/
	    case ST_PUTDATA:
		request->xfrstat += (count_t)recvLength;
		recvLength -= sizeof(tftpH_t);

		/* compuration below may be wrong for netascii transfers */

                if (request->maxsize &&
                    ((recvLength + request->iostat) > request->maxsize)) {
                    LOG_MSG2 (LOG_WARNING, 
                              _("Data exceeds ACL limit: %ld bytes"),
                              request->maxsize);
                    sendError(request, PE_NOSPACE, NULL);
                    return -1;
                }
                if (request->option.totalsize) {
                    if ((recvLength+request->iostat)>request->option.totalsize){
                        LOG_MSG2 (LOG_WARNING, 
                                  _("Negotiated data exceeds %ld bytes"),
                                  request->option.totalsize);
                        sendError(request, PE_NOSPACE, NULL);
                        return -1;
                    }
                }   
		done = (recvLength < request->option.blocksize);
		if (putXferBuffer (request->reqFd, recvBuffer->data, 
		                   recvLength, request->reqMode, done,
                                   &request->iostat) != recvLength) {
                    LOG_MSG1 (LOG_ERR, _("Cannot write local data: %m"));
                    sendError(request, PE_NOSPACE, NULL);
                    return -1;
		}
		LOG_MSG6 (LOG_DEBUG, "%s: %d bytes to file %s",__FUNCTION__, 
                                     recvLength, 
                                     done ? " (last block)" : "");
		state = ST_ACKWAIT;
		request->curBlock++;
		request->retry = 0;
		break;
/*
	wait for send channel to become ready
	we don't tolerate timeouts at this point
	on success, proceed with 'send'
*/
	    case ST_ACKWAIT:
		FD_ZERO(&fdSel);
		FD_SET(request->Ssock, &fdSel);

		timeout = proBackoff(&request->times, EBS_SACK, request->retry);

                LOG_MSG6 (LOG_DEBUG, "%s: timeout %.1fs", __FUNCTION__,
                          tv2double(&timeout));

		rtc = select (hsock, NULL, &fdSel, NULL, &timeout);

                switch(checkSelError(request, rtc, BM_ACK)) {
                    case PC_ERROR:
                        return -1;
                    case PC_AGAIN:
                    case PC_RETRY:
                        continue;
                }
		state = ST_ACK;
		break;
/*
	send an ACK for the block
*/
	    case ST_ACK:
		ackBuffer.head.opCode  = htons(RT_ACK);
		ackBuffer.head.blockNo = htons(lastDat);

		rtc = send(request->Ssock, &ackBuffer, sizeof(tftpH_t), 0);

                switch(checkSendError(request, rtc, sizeof(tftpH_t), BM_ACK)) {
                    case PC_ERROR:
		        return -1;
                    case PC_RETRY:
		        state = ST_ACKWAIT;
                        continue;
		}

		LOG_MSG6 (LOG_DEBUG, "%s: ACK for block %d%s", __FUNCTION__,
			  lastDat, lastDat ? "": " (init transfer)");

		if (done) {
		    return 0;
		}

		state = ST_RECVWAIT;
		break;
/*
	wait for send channel to become ready
	we don't tolerate timeouts at this point
	on success, proceed with OACK
*/
	    case ST_OACKWAIT:
		FD_ZERO(&fdSel);
		FD_SET(request->Ssock, &fdSel);

		timeout = proBackoff(&request->times, EBS_SACK, request->retry);

                LOG_MSG6 (LOG_DEBUG, "%s: timeout %.1fs", __FUNCTION__,
                          tv2double(&timeout));

		rtc = select (hsock, NULL, &fdSel, NULL, &timeout);

                switch(checkSelError(request, rtc, BM_OACK)) {
                    case PC_ERROR:
                        return -1;
                    case PC_AGAIN:
                    case PC_RETRY:
                        continue;
                }
		state = ST_OACK;
		break;
/*
	send an OACK for the option request
*/
	    case ST_OACK:
		ackBuffer.head.opCode  = htons(RT_OACK);
		ackBuffer.head.blockNo = htons(0);

		LOG_MSG6 (LOG_DEBUG, "Acknowledged options: %s",
                          textOptions(request));

                plen = sizeof(ackBuffer.data);
                appendOptions(ackBuffer.data, request, &plen);

                if (plen < 0) {
                    sendError(request, PE_UNDEF, "Internal server error");
		    return -1;
                }
                else
                    plen += sizeof(ackBuffer.head);

		rtc = send(request->Ssock, &ackBuffer, plen, 0);

                switch(checkSendError(request, rtc, plen, BM_OACK)) {
                    case PC_ERROR:
		        return -1;
                    case PC_RETRY:
		        state = ST_OACKWAIT;
                        continue;
		}
		state = ST_RECVWAIT;
		request->retry = 0;
		break;
	}
    }
    return -1;
}


/*
	read request (RRQ): send a file to the client
	we assume an already opened file and an active socket
	all security checks  re done and 
	options (if applicable) are processed
	mode may be one of FM_OCTET or FM_NETASCII

	jacket routine for sendFile2 (the real thing, see below)
	this is to assure an easy cleanup of dyn memory at function return
*/
int sendFile (tftpR_t *request)
{
    tftpD_t *sendBuffer;	/* send buffer (one packet)	*/
    int rtc, ios, bufsize;

    if ((request->option.types & ROT_TMO)) {
        request->times.recv = request->times.sack = request->option.timeout;
        LOG_MSG4(LOG_INFO, _("New Timing: %s"), 
                             timer2asc(&request->times, T2A_XFER|T2A_RET));
    }
    LOG_MSG5 (LOG_DEBUG, _("Progressive backoff: %s"),
                           listBackoff(&request->times, EBS_SACK));

    request->times.start = request->times.stop  = uftime(NULL);

    LOG_MSG5 (LOG_DEBUG, "Blocksize: %d %s", request->option.blocksize,
                        (request->option.types & ROT_BSZ) ? 
                            "(requested by client)" : "");

    bufsize = MAX(request->option.blocksize, SEGSIZE) + sizeof(tftpH_t);

    if ((sendBuffer = (tftpD_t *)Malloc(bufsize)) == NULL) {
        errNoDynmem;
        request->times.stop = uftime(NULL);
        return -1;
    }

#ifdef FILE_CACHE
    if (request->fcache != NULL)
        ios = initCacheBuffer(request->fcache);
    else
#endif /* FILE_CACHE */
        ios = initXferBuffer(request->option.totalsize);

    if (ios) {
        sendError(request, PE_NOSPACE, NULL);
        return -1;
    }

    rtc = sendFile2 (request, sendBuffer);

#ifdef FILE_CACHE
/**/LOG_MSG3(LOG_NOTICE, "Data load time: %g s", tv2double(&getSum));
    if (request->fcache == NULL)
#endif /* FILE_CACHE */
        rmXferBuffer();

    request->times.stop = uftime(NULL);
    Free(sendBuffer);
    return rtc;
}

/*
	transfer processing according to RFC 2347:

	client                                           server
	-------------------------------------------------------
	|1|foofile|0|octet|0|blksize|0|1432|0|  -->         RRQ
	                       <--  |6|blksize|0|1432|0|   OACK
	|4|0|  -->                                          ACK
	                <--  |3|1| 1432 octets of data |   DATA
	|4|2|  -->                                          ACK
	                <--  |3|3|<1432 octets of data |   DATA
	|4|3|  -->                                          ACK
*/

static int sendFile2 (tftpR_t *request, tftpD_t *sendBuffer)
{
    tftpD_t ackBuffer;		/* acknowledge packet		*/
    int recvLength;		/* received bytes		*/
    int sendLength = 0;		/* sent bytes			*/
    int bufLength = 0;		/* lenght of transfer buffer	*/
    int hsock;			/* highest socket+1 (select)	*/
    ushort thisAck = 0;		/* currently ack ACK'ed block	*/
    int done = 0;		/* flag for end-of-transfer	*/
    int rtc;			/* general function return code	*/
    fd_set fdSel;		/* select bitmap		*/
    int plen;			/* length of OACK packet	*/
    enum io_states state;	/* state machine status		*/
    struct timeval timeout;	/* i/o timer			*/

    hsock = MAX(request->Rsock, request->Ssock) + 1;
    request->curBlock = 0;		/* netxt block awaited		*/

#ifdef FILE_CACHE
/**/memset(&getSum,0,sizeof(struct timeval));
#endif /* FILE_CACHE */

/*
	Loop through the state-machine that implements the send algorithm

	whithout options (none requested or not enabled explicitely)
	we should start sending the first data packet.

	with option processing there is annother cycle of acknowlegement
	needed: we continue with sending an OACK and wait for the ACK.
*/
    state = (xferOptions && request->option.types) ? ST_OACKWAIT : ST_GETDATA;

    for (;;) {

        if (doTrace) {
            doTrace = 0;
            statusTrace (request);
        }

        LOG_MSG6 (LOG_DEBUG, "%s: state %d: %s", __FUNCTION__,
                  state,state_names[state]);

        switch (state) {

            default:
                LOG_MSG1 (LOG_ERR, _("Internal state machine error"));
                break;
/*
	wait for send channel to become ready
	we don't tolerate timeouts at this point
	on success, proceed with OACK
*/
            case ST_OACKWAIT:
                FD_ZERO(&fdSel);
                FD_SET(request->Ssock, &fdSel);

		timeout = proBackoff(&request->times, EBS_SEND, request->retry);

                LOG_MSG6 (LOG_DEBUG, "%s: timeout %.1fs", __FUNCTION__,
                          tv2double(&timeout));

                rtc = select (hsock, NULL, &fdSel, NULL, &timeout);

                switch(checkSelError(request, rtc, BM_OACK)) {
                    case PC_ERROR:
                        return -1;
                    case PC_AGAIN:
                    case PC_RETRY:
                        continue;
                }
                state = ST_OACK;
                break;
/*
	build & send an option acknowlegement packet to client
	the OACK data field is essentially the same that was requested
	from the client with following exceptions:
		- filename and transfer mode omitted
		- blocksize adjusted (downwards) to our needs
		- total size filled in with the file size
*/
            case ST_OACK:
		ackBuffer.head.opCode  = htons(RT_OACK);
		ackBuffer.head.blockNo = htons(0);

		LOG_MSG6 (LOG_DEBUG, "Acknowledged options: %s",
                          textOptions(request));

                plen = sizeof(ackBuffer.data);
                appendOptions(ackBuffer.data, request, &plen);

                if (plen < 0) {
                    sendError(request, PE_UNDEF, "Server options problem");
                    return -1;
                }
                else
                    plen += sizeof(ackBuffer.head);

		rtc = send(request->Ssock, &ackBuffer, plen, 0);

                switch(checkSendError(request, rtc, plen, BM_OACK)) {
                    case PC_ERROR:
		        return -1;
                    case PC_RETRY:
		        state = ST_OACKWAIT;
                        continue;
		}

		LOG_MSG6 (LOG_DEBUG, "%s: OACK", __FUNCTION__);

		state = ST_ACKWAIT;
		request->retry = 0;
                break;

/*
	get the next chunk of data from disk
	on success fall through to 'send wait'
*/ 
            case ST_GETDATA:
                while (request->curBlock <= request->lastAck) {
#ifdef FILE_CACHE
/**/                struct timeval tStart, tEnd;
/**/                tStart = uftime(NULL);
                    if (request->fcache != NULL) {
                        bufLength = getCacheBuffer(sendBuffer->data, 
                                                   request->option.blocksize);
                        request->iostat += bufLength;
                    }
                    else
#endif /* FILE_CACHE */
                        bufLength = getXferBuffer (request->reqFd,
                                                   sendBuffer->data, 
                                                   request->option.blocksize, 
                                                   request->reqMode,
                                                   &request->iostat);
#ifdef FILE_CACHE
/**/                tEnd = uftime(&tStart);
/**/                getSum = tvSum(getSum,tEnd);
#endif /* FILE_CACHE */
                    
                    if (bufLength < 0) {
                        LOG_MSG1 (LOG_ERR, _("Cannot read local data: %m"));
                        sendError(request, PE_UNDEF, "No data");
                        return -1;
                    }
                    if (hs(request->curBlock) > hs(request->curBlock+1))
                        LOG_MSG5 (LOG_DEBUG, "Sequence wrap at block %d",
                                  request->curBlock+1);
                    request->curBlock++;
                    if (bufLength < request->option.blocksize) {
                        done = 1;
                        break;
                    }
                }
                if (request->curBlock <= request->lastAck) {
                    LOG_MSG1 (LOG_ERR, _("Attempt to skip over EOF"));
                    sendError(request, PE_ACCESS, "Attempt to skip over EOF");
/*
	NOTE: systematic error possible
	if the received 'step-forward-ack' blok # was messed by a transmission
	error, we're in trouble: allowing a retry and then getting a lower
	block# leads us in a non recoverable state. thus, for now we
	don't allow retries right here !
*/
                    return -1;
                }        
                sendBuffer->head.opCode  = htons(RT_DATA);
                sendBuffer->head.blockNo = htons(hs(request->curBlock));
                sendLength = sizeof(tftpH_t) + bufLength;
                request->retry = 0;

                LOG_MSG6 (LOG_DEBUG, "%s: %d bytes from file", __FUNCTION__, 
                          bufLength);

                state = ST_SENDWAIT;
                break;
/*
	wait for send channel to become ready
	we don't tolerate timeouts at this point
	on success, proceed with SEND
*/
            case ST_SENDWAIT:
                FD_ZERO(&fdSel);
                FD_SET(request->Ssock, &fdSel);

		timeout = proBackoff(&request->times, EBS_SEND, request->retry);

                LOG_MSG6 (LOG_DEBUG, "%s: timeout %.1fs", __FUNCTION__,
                          tv2double(&timeout));

                rtc = select (hsock, NULL, &fdSel, NULL, &timeout);

                switch (checkSelError(request, rtc, BM_DATA)) {
                    case PC_ERROR:
                        return -1;
                    case PC_AGAIN:
                    case PC_RETRY:
                         continue;
                    default:
                         state = ST_SEND;
                }
                break;
/*
	send the current buffer
*/
            case ST_SEND:

                rtc = send(request->Ssock, sendBuffer, sendLength, 0);

                switch(checkSendError(request, rtc, sendLength, BM_DATA)) {
                    case PC_ERROR:
		        return -1;
                    case PC_RETRY:
		        state = ST_SENDWAIT;
                        continue;
		}

                LOG_MSG6 (LOG_DEBUG, "%s: block %d, size %d sent (%d)",
                          __FUNCTION__, request->curBlock, sendLength,
                          request->retry+1); 
                state = ST_ACKWAIT;
                break;
/*
	wait for ACK from peer
	this is actually a 'wait for any reply',
	because we shoud be prepared to receive an packet,
	especially error packets.
*/
            case ST_ACKWAIT:
                FD_ZERO(&fdSel);
                FD_SET(request->Rsock, &fdSel);

		timeout = proBackoff(&request->times, EBS_SACK, request->retry);

                LOG_MSG6 (LOG_DEBUG, "%s: timeout %.1fs", __FUNCTION__,
                          tv2double(&timeout));

                rtc = select (hsock, &fdSel, NULL, NULL, &timeout);

                switch(checkSelError(request, rtc, BM_ACK)) {
                    case PC_ERROR:
                        return -1;
                    case PC_AGAIN:
                        continue;
                    case PC_RETRY:
                        state = request->curBlock ? ST_SENDWAIT : ST_OACKWAIT;
                        break;
                    default:
                        state = ST_ACK;
                }
                break;
/*
	The received packed should be at least of length sizeof(tftpD_t)
	(usually 4 bytes). Short packets are silently discarded:
	we proceed as with timeouts above.

	ACK packages coming from the wrong source port will not reach us,
	because of previous connect()
*/
            case ST_ACK:
                recvLength = recv(request->Rsock, &ackBuffer, 
                                  sizeof(tftpH_t) + SEGSIZE, MSG_TRUNC);
                switch (checkAck(request, &ackBuffer, recvLength)) {
                    case PC_ERROR:
                        return -1;
                    case PC_RETRY:
                        state = ST_SENDWAIT;
                        break;
                }

/*
	We got the desired ACK packet, so lets check the block number.

	Warning: The Block number is 16-bit unsiggned an thus wrappes
	from 65535 to 0 ! The internal block counter (curBlock) reflects
	the REAL (32-bit) block number and is mapped back to uint16
	using the hs() macro.

	Valid cases:	
		current-block:	The latest packet sent was ACK'ed.
				We may proceed with the next one.
		previous-block:	The latest packet was lost or
				destroyed, thus not ACK'ed.

                                With altenate protocol (RFC1123):
				The ACK may com in delayed, while DATA
				has already been sent twice (or more).
				In this case the repeat count is > 0 !

	Without recovery enabled (normal case):

		All other cases are considered transmission errors and
		will take the retransmission path.

	Recovery enabled (forward spacing allowed):

		The block number is considered correct and reflects
		the latest packet the client reclaims o.k.	

	Note: with option processing this can be either an acknowlege
	of a data block (real block# > 0) or the acknowlege of an OACK packet
	in which case the real block# is 0.
*/
                thisAck = ntohs(ackBuffer.head.blockNo);

                LOG_MSG6 (LOG_DEBUG, "%s: got ACK for block %hu",
                          __FUNCTION__, thisAck);

                if (thisAck == hs(request->curBlock)) {
                    request->xfrstat += (count_t)sendLength;
                    if (done) 
                        return 0;
                    request->lastAck = request->curBlock;
                    state = ST_GETDATA;
                }

/*
	this section deals with the protocol modification to prevent
	the ping-pong duplication following a delayed ACK
*/
                else if ((thisAck == hs(request->lastAck)) &&
                         !(request->times.eflag & SD_OLDPROTO)) {
                    LOG_MSG6 (LOG_DEBUG, 
                              "%s: dup ACK: new proto - data no resent",
                              __FUNCTION__);
                    if (retryPossible(request,PE_UNDEF,NULL)!=PC_RETRY)
                        return -1;
                    state = ST_ACKWAIT;
                }
/*
	here if old protocol handling is explicitely selected
	and in all other cases of non matching sequence numbers
*/
                else {
                    if (!allowRecover || done) {
                        logSequenceError("ACK", (int)thisAck,
                                         (int)hs(request->curBlock));
                        if (retryPossible(request, PE_UNDEF, "Sequence Error")
                            != PC_RETRY)
                            return -1;
                        state = ST_SENDWAIT;
                    }
                    else {
                        request->lastAck += (int)(thisAck-hs(request->lastAck));
                        if(thisAck < hs(request->lastAck)) /* rollover */
                            request->lastAck += USHRT_MAX+1;

                        LOG_MSG4 (LOG_INFO, _("Step from %d to %d"),
                                  request->curBlock, request->lastAck+1);
                        state = ST_GETDATA;
                    }
                }
                break;
        } 
    }
    return -1;
}
